@@ -12,6 +12,7 @@ from typing import Dict, Iterable, List, Optional, Union
 import requests
 
 # first party
+from dbtc import models
 from dbtc.client.base import _Client
 from dbtc.utils import json_listify, listify
 
@@ -49,11 +50,33 @@ SUB_COMMAND_CLI_ARGS = {
 }
 
 
+def set_called_from(func):
+    @wraps(func)
+    def wrapper(self, *args, **kwargs):
+
+        # Don't want to reset this when it already exists
+        if self._called_from is None:
+            self._called_from = func.__name__
+        try:
+            result = func(self, *args, **kwargs)
+        except Exception as e:
+            self._called_from = None
+            raise (e)
+
+        self._called_from = None
+        return result
+
+    return wrapper
+
+
 def _version_decorator(func, version):
     @wraps(func)
     def wrapper(self, *args, **kwargs):
         self._path = f'/api/{version}/'
         result = func(self, *args, **kwargs)
+        if not self.do_not_track:
+            self._send_track('Admin API', func, *args, **kwargs)
+
         return result
 
     return wrapper
@@ -61,7 +84,6 @@ def _version_decorator(func, version):
 
 v2 = partial(_version_decorator, version='v2')
 v3 = partial(_version_decorator, version='v3')
-v4 = partial(_version_decorator, version='v4')
 
 
 class _AdminClient(_Client):
@@ -91,6 +113,14 @@ class _AdminClient(_Client):
         self, path: str, *, method: str = 'get', **kwargs
     ) -> requests.Response:
         """Make request to API."""
+
+        # Model is not an argument that the request method accepts, needs to be removed
+        model = kwargs.pop('model', None)
+        if model is not None:
+
+            # This will validate the payload
+            kwargs['json'] = model(**kwargs['json']).dict(exclude_unset=True)
+
         full_url = self.full_url(path)
         response = self.session.request(method=method, url=full_url, **kwargs)
         return response
@@ -100,39 +130,6 @@ class _AdminClient(_Client):
         response = self._make_request(path, method=method, **kwargs)
         return response.json()
 
-    def _paginated_request(
-        self, path: str, *, method: str = 'get', **kwargs
-    ) -> List[Dict]:
-        """Multiple paginated requests given presence of specific header.
-
-        !!! note
-            Only available in V4.
-        """
-        response = self._make_request(path, method=method, **kwargs)
-        data = []
-        while True:
-            response_data = response.json().get('data', [])
-            data.extend(response_data)
-            next_page_token = self._get_pagination_token(response)
-            if next_page_token is not None:
-                response = self._make_request(
-                    path,
-                    method=method,
-                    headers={'x-dbt-continuation-token': next_page_token},
-                    **kwargs,
-                )
-            else:
-                break
-        return data
-
-    def _get_pagination_token(self, response):
-        """Retrieve pagination token.
-
-        !!! note
-            Only available in V4.
-        """
-        return response.headers.get('x-dbt-continuation-token', None)
-
     def _get_by_name(self, items: List, item_name: str, value: str = 'name'):
         try:
             obj = [item for item in items if item[value] == item_name][0]
@@ -360,6 +357,21 @@ class _AdminClient(_Client):
             f'accounts/{account_id}/service-tokens/', method='post', json=payload
         )
 
+    @v3
+    def create_webhook(self, account_id: int, payload: Dict) -> Dict:
+        """Create a new outbound webhook
+
+        Args:
+            account_id (int): Numeric ID of the account
+            payload (dict): Dictionary representing the webhook to create
+        """
+        return self._simple_request(
+            f'accounts/{account_id}/webhooks/subscriptions',
+            method='post',
+            json=payload,
+            model=models.Webhook,
+        )
+
     @v3
     def create_user_group(self, account_id: int, payload: Dict) -> Dict:
         """Create a user group
@@ -487,6 +499,19 @@ class _AdminClient(_Client):
             method='delete',
         )
 
+    @v3
+    def delete_webhook(self, account_id: int, webhook_id: str) -> Dict:
+        """Delete a webhook
+
+        Args:
+            account_id (int): Numeric ID of the account
+            webhook_id (str): String ID of the webhook you want to delete
+        """
+        return self._simple_request(
+            f'accounts/{account_id}/webhooks/subscription/{webhook_id}',
+            method='delete',
+        )
+
     @v3
     def delete_user_group(self, account_id: int, group_id: int) -> Dict:
         """Delete group for a specified account
@@ -509,6 +534,7 @@ class _AdminClient(_Client):
         """
         return self._simple_request(f'accounts/{account_id}')
 
+    @set_called_from
     @v2
     def get_account_by_name(self, account_name: str) -> Dict:
         """Get an account by its name.
@@ -557,6 +583,7 @@ class _AdminClient(_Client):
         """
         return self._simple_request(f'accounts/{account_id}/projects/{project_id}')
 
+    @set_called_from
     @v2
     def get_project_by_name(
         self, project_name: str, account_id: int = None, account_name: str = None
@@ -592,6 +619,7 @@ class _AdminClient(_Client):
 
         raise Exception(f'Project "{project_name}" was not found.')
 
+    @set_called_from
     @v2
     def get_most_recent_run(
         self,
@@ -604,6 +632,22 @@ class _AdminClient(_Client):
         deferring_run_id: int = None,
         status: Union[List[str], str] = None,
     ) -> Dict:
+        """Get the most recent run
+
+        Args:
+            account_id (int): Numeric ID of the account to retrieve
+            include_related (list): List of related
+                fields to pull with the run. Valid values are `trigger`, `job`,
+                `repository`, `debug_logs`, `run_steps`, and `environment`.
+            job_definition_id (int, optional): Applies a filter to only return
+                runs from the specified Job.
+            environment_id (int, optional): Numeric ID of the environment
+            project_id (int or list, optional): The project ID or IDs
+            deferring_run_id (int, optional): Numeric ID of a deferred run
+            status (str or list, optional): The status to apply when listing runs.
+                Options include queued, starting, running, success, error, and
+                cancelled
+        """
         runs = self.list_runs(
             account_id,
             include_related=include_related,
@@ -621,6 +665,7 @@ class _AdminClient(_Client):
             runs['data'] = {}
         return runs
 
+    @set_called_from
     @v2
     def get_most_recent_run_artifact(
         self,
@@ -633,6 +678,37 @@ class _AdminClient(_Client):
         deferring_run_id: int = None,
         step: int = None,
     ):
+        """Fetch artifacts from the most recent run
+
+        Once a run has completed, you can use this endpoint to download the
+        `manifest.json`, `run_results.json` or `catalog.json` files from dbt Cloud.
+        These artifacts contain information about the models in your dbt project,
+        timing information around their execution, and a status message indicating the
+        result of the model build.
+
+        !!! note
+            By default, this endpoint returns artifacts from the last step in the
+            run. To list artifacts from other steps in the run, use the step query
+            parameter described below.
+
+        !!! warning
+            If requesting a non JSON artifact, the result will be a `str`
+
+        Args:
+            account_id (int): Numeric ID of the account to retrieve
+            path (str): Paths are rooted at the target/ directory. Use manifest.json,
+                catalog.json, or run_results.json to download dbt-generated artifacts
+                for the run.
+            job_definition_id (int, optional): Applies a filter to only return
+                runs from the specified Job.
+            environment_id (int, optional): Numeric ID of the environment
+            project_id (int or list, optional): The project ID or IDs
+            deferring_run_id (int, optional): Numeric ID of a deferred run
+            step (str, optional): The index of the Step in the Run to query for
+                artifacts. The first step in the run has the index 1. If the step
+                parameter is omitted, then this endpoint will return the artifacts
+                compiled for the last step in the run.
+        """
         runs = self.get_most_recent_run(
             account_id,
             job_definition_id=job_definition_id,
@@ -641,6 +717,10 @@ class _AdminClient(_Client):
             deferring_run_id=deferring_run_id,
             status='success',
         )
+
+        # Reset called from after being set to None in get_most_recent_run
+        self._called_from = 'get_most_recent_run_artifact'
+
         try:
             run_id = runs.get('data', {})['id']
         except KeyError:
@@ -724,16 +804,6 @@ class _AdminClient(_Client):
             f'accounts/{account_id}/projects/{project_id}/runs/{run_id}/timing/'
         )
 
-    @v4
-    def get_run_v4(self, account_id: int, run_id: int) -> Dict:
-        """Retrieves the details of an existing run with the given run_id.
-
-        Args:
-            account_id (int): Numeric ID of the account to retrieve
-            run_id (int): Numeric ID of the run to retrieve
-        """
-        return self._simple_request(f'accounts/{account_id}/runs/{run_id}')
-
     @v3
     def get_service_token(self, account_id: int, service_token_id: int) -> Dict:
         """Retrieves a service token.
@@ -746,6 +816,18 @@ class _AdminClient(_Client):
             f'accounts/{account_id}/service-tokens/{service_token_id}'
         )
 
+    @v3
+    def get_webhook(self, account_id: int, webhook_id: str) -> Dict:
+        """Get a webhook
+
+        Args:
+            account_id (int): Numeric ID of the account
+            webhook_id (str): String ID of the webhook you want to retrieve
+        """
+        return self._simple_request(
+            f'accounts/{account_id}/webhooks/subscription/{webhook_id}',
+        )
+
     @v2
     def get_user(self, account_id: int, user_id: int) -> Dict:
         """List invited users in an account.
@@ -846,7 +928,7 @@ class _AdminClient(_Client):
 
         Args:
             account_id (int): Numeric ID of the account to retrieve
-            project_id (int, optional): Numeric ID of the project to retrieve
+            project_id (int or list, optional): The project ID or IDs
             dbt_version (str or list, optional): The version of dbt the environment
                 is using
             name (str, optional): Name of the environment to retrieve
@@ -1011,7 +1093,7 @@ class _AdminClient(_Client):
         include_related: List[str] = None,
         job_definition_id: int = None,
         environment_id: int = None,
-        project_id: int = None,
+        project_id: Union[int, List[int]] = None,
         deferring_run_id: int = None,
         status: Union[List[str], str] = None,
         order_by: str = None,
@@ -1026,17 +1108,19 @@ class _AdminClient(_Client):
                 fields to pull with the run. Valid values are `trigger`, `job`,
                 `repository`, `debug_logs`, `run_steps`, and `environment`.
             job_definition_id (int, optional): Applies a filter to only return
-                runs
-                from the specified Job.
+                runs from the specified Job.
+            environment_id (int, optional): Numeric ID of the environment
+            project_id (int or list, optional): The project ID or IDs
+            deferring_run_id (int, optional): Numeric ID of a deferred run
+            status (str or list, optional): The status to apply when listing runs.
+                Options include queued, starting, running, success, error, and
+                cancelled
             order_by (str, optional): Field to order the result by.
                 Use - to indicate reverse order.
             offset (int, optional): The offset to apply when listing runs.
                 Use with limit to paginate results.
             limit (int, optional): The limit to apply when listing runs.
                 Use with offset to paginate results.
-            status (str or list, optional): The status to apply when listing runs.
-                Options include queued, starting, running, success, error, and
-                cancelled
         """
         if status is not None:
             try:
@@ -1060,44 +1144,6 @@ class _AdminClient(_Client):
             },
         )
 
-    @v4
-    def list_runs_v4(
-        self,
-        account_id: int,
-        *,
-        limit: int = None,
-        environment: str = None,
-        project: str = None,
-        job: str = None,
-        status: str = None,
-    ) -> List[Dict]:
-        """Returns a list of runs in the account.
-
-        The runs are returned sorted by creation date, with the most recent run
-        appearing first.
-
-        Args:
-            account_id (int): Numeric ID of the account to retrieve
-            limit (int, optional): A limit on the number of objects to be
-                returned, between 1 and 100.
-            environment (str): A filter on the list based on the object's
-                environment_id field.
-            project (str): A filter on the list based on the object's project_id field.
-            job (str): A filter on the list based on the object's job_id field.
-            status: A filter on the list based on the object's status field.
-                Enum: "Queued" "Starting" "Running" "Succeeded" "Failed" "Canceled"
-        """
-        return self._paginated_request(
-            f'accounts/{account_id}/runs',
-            params={
-                'limit': limit,
-                'environment': environment,
-                'project': project,
-                'job': job,
-                'status': status,
-            },
-        )
-
     @v3
     def list_service_token_permissions(
         self, account_id: int, service_token_id: int
@@ -1121,11 +1167,12 @@ class _AdminClient(_Client):
         """
         return self._simple_request(f'accounts/{account_id}/service-tokens/')
 
-    @v2
+    @v3
     def list_users(
         self,
         account_id: int,
         *,
+        state: int = None,
         limit: int = None,
         offset: int = None,
         order_by: str = 'email',
@@ -1134,6 +1181,7 @@ class _AdminClient(_Client):
 
         Args:
             account_id (int): Numeric ID of the account to retrieve
+            state (int, optional): 1 = active, 2 = deleted
             limit (int, optional): The limit to apply when listing runs.
                 Use with offset to paginate results.
             offset (int, optional): The offset to apply when listing runs.
@@ -1143,7 +1191,33 @@ class _AdminClient(_Client):
         """
         return self._simple_request(
             f'accounts/{account_id}/users/',
-            params={'limit': limit, 'offset': offset, 'order_by': order_by},
+            params={
+                'limit': limit,
+                'offset': offset,
+                'order_by': order_by,
+                'state': state,
+            },
+        )
+
+    @v3
+    def list_webhooks(
+        self,
+        account_id: int,
+        *,
+        limit: int = None,
+        offset: int = None,
+    ) -> Dict:
+        """List of webhooks in account
+        Args:
+            account_id (int): Numeric ID of the account
+            limit (int, optional): The limit to apply when listing runs.
+                Use with offset to paginate results.
+            offset (int, optional): The offset to apply when listing runs.
+                Use with limit to paginate results.
+        """
+        return self._simple_request(
+            f'accounts/{account_id}/webhooks/subscriptions',
+            params={'limit': limit, 'offset': offset},
         )
 
     @v3
@@ -1158,6 +1232,19 @@ class _AdminClient(_Client):
             f'accounts/{account_id}/connections/test/', method='post', json=payload
         )
 
+    @v3
+    def test_webhook(self, account_id: int, webhook_id: str) -> Dict:
+        """Test a webhook
+
+        Args:
+            account_id (int): Numeric ID of the account
+            webhook_id (str): String ID of the webhook you want to test
+        """
+        return self._simple_request(
+            f'accounts/{account_id}/webhooks/subscription/{webhook_id}/test',
+        )
+
+    @set_called_from
     @v2
     def trigger_autoscaling_ci_job(
         self,
@@ -1168,6 +1255,7 @@ class _AdminClient(_Client):
         should_poll: bool = False,
         poll_interval: int = 10,
         delete_cloned_job: bool = True,
+        max_run_slots: int = None,
     ):
         """Trigger an autoscaling CI job
 
@@ -1202,6 +1290,11 @@ class _AdminClient(_Client):
                 polling
             delete_cloned_job (bool, optional): Indicate if cloned job should be
                 deleted after being triggered
+            max_run_slots (int, optional): Number of run slots that should be
+                available to this process.  This will limit the ability to run
+                concurrent PRs up the the allocated run slots for your account.  When
+                set to `None`, the `run_slots` allocated to your account will be used
+                to determine if a job should be cloned.
         """
         self.console.log('Finding any in progress runs...')
         cloned_job = None
@@ -1260,9 +1353,11 @@ class _AdminClient(_Client):
             # cancelled above
             if not job_run_is_pr_run:
 
-                acct = self.get_account(account_id).get('data', {})
-                run_slots = acct.get('run_slots', 0)
-                if run_slots > len(in_progress_runs):
+                run_slots = (
+                    self.get_account(account_id).get('data', {}).get('run_slots', 0)
+                )
+                max_run_slots = min(max_run_slots or run_slots, run_slots)
+                if max_run_slots > len(in_progress_runs):
                     self.console.log(
                         f'Job {job_id} is currently being used in run {job_run["id"]}. '
                         'This job definition will be cloned and then triggered for '
@@ -1271,7 +1366,9 @@ class _AdminClient(_Client):
                     current_job = self.get_job(account_id, job_id).get('data', {})
 
                     # Alter the current job definition so it can be cloned
-                    current_job.pop('is_deferrable')
+                    read_only_fields = ['is_deferrable', 'raw_dbt_version']
+                    for read_only_field in read_only_fields:
+                        current_job.pop(read_only_field)
                     current_job['id'] = None
                     now = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
                     current_job['name'] = current_job['name'] + f' [CLONED {now}]'
@@ -1286,8 +1383,8 @@ class _AdminClient(_Client):
                 else:
                     self.console.log(
                         'Not cloning the job as your account has met or exceeded the '
-                        'number of run slots and will not be able to execute even a '
-                        'cloned CI job.'
+                        'number of run slots or a limit was placed by the user.  The '
+                        'normal job will be queued.'
                     )
         else:
             self.console.log('No in progress job run found.  Triggering as normal')
@@ -1298,10 +1395,14 @@ class _AdminClient(_Client):
             should_poll=should_poll,
             poll_interval=poll_interval,
         )
+
+        # This property was set to `None` in the trigger_job method, reset here
+        self._called_from = 'trigger_autoscaling_ci_job'
         if cloned_job is not None and delete_cloned_job:
             self.delete_job(account_id, job_id)
         return run
 
+    @set_called_from
     @v2
     def trigger_job_from_failure(
         self,
@@ -1310,7 +1411,7 @@ class _AdminClient(_Client):
         payload: Dict,
         *,
         should_poll: bool = True,
-        poll_interval: bool = False,
+        poll_interval: int = 10,
         trigger_on_failure_only: bool = True,
     ):
         """Restart a job from the point of failure
@@ -1450,7 +1551,7 @@ class _AdminClient(_Client):
             if trigger_on_failure_only:
                 self.console.log('Not triggering job because prior run was successful.')
                 return
-        return self.trigger_job(
+        run = self.trigger_job(
             account_id,
             job_id,
             payload,
@@ -1458,6 +1559,11 @@ class _AdminClient(_Client):
             poll_interval=poll_interval,
         )
 
+        # This property was set to `None` in the trigger_job method, reset here
+        self._called_from = 'trigger_job_from_failure'
+        return run
+
+    @set_called_from
     @v2
     def trigger_job(
         self,
@@ -1619,3 +1725,18 @@ class _AdminClient(_Client):
             method='post',
             json=payload,
         )
+
+    @v3
+    def update_webhook(self, account_id: int, webhook_id: str, payload: Dict) -> Dict:
+        """Update a webhook
+
+        Args:
+            account_id (int): Numeric ID of the account
+            webhook_id (str): String ID of the webhook you want to update
+            payload (dict): Dictionary representing the webhook to update
+        """
+        return self._simple_request(
+            f'accounts/{account_id}/webhooks/subscription/{webhook_id}',
+            method='put',
+            json=payload,
+        )
